/*	FIGLIST
 *
 *	Displays all installed auto-configured expansion boards.
 *
 *	Usage:	CLI>  figlist  [ ? ]
 *
 *	where:
 *
 *		- no argument causes the present system's ConfigList to be
 *		  displayed
 *		- any argument causes a brief description to be displayed
 *		  along with the known and unknown manufacturer codes.
 *
 *	Thad Floryan, 2-Sept-1988
 *
 *	If you have updates to the manufacturers or codes, please notify:
 *
 *	UUCP:	thad@cup.portal.com (OR) ...!sun!portal!cup.portal.com!thad
 *	UUCP:	thad@well.UUCP
 *	BIX:	tfloryan
 *	DELPHI:	THADF
 *	BBS:  *	BBS-HT,   408/737-0900 (300,1200)      "Thad Floryan" or "SYSOP"
 *	BBS:	BBS-JC,   415/961-7250 (300,1200,2400) "Thad Floryan" or "SYSOP"
 *	BBS:	FAUG-BBS, 415/595-2479 (300,1200,2400) "Thad Floryan"
 *	BBS:	Amiga OZ, 316/283-9210 (300,1200,2400) "Thad Floryan" or "SYSOP"
 *
 *	      * = store BBS, California 6PM-8AM Mon-Fri, 5PM Sat-8AM Mon
 *
 *	Compile/link using Manx per:
 *
 *		cc figlist
 *		ln figlist -lc
 */

#include <exec/libraries.h>
#include <libraries/configvars.h>
#include <libraries/expansion.h>

extern void		CloseLibrary();
extern struct ConfigDev	*FindConfigDev();
extern struct Library	*OpenLibrary();

struct Library		*ExpansionBase;

static char *boardmemsize[] = {
    "8MB", "64KB", "128KB", "256KB", "512KB", "1MB", "2MB", "4MB"};

struct MfrCodes {
	UWORD	mc_Code;
	char  * mc_Name;
} MfrTable[] = {
	0x0202, "Commodore",
	0x03ec, "Pacific Peripherals",
	0x03ee, "ComSpec Communications",
	0x03f6, "Expansion Technologies (Pacific Cypress)",
	0x0420, "Supra Corporation",
	0x07e5, "Synergy",
/*
 *	codes for the following are unknown
 */
	0x0000, "A-Squared",			/* Live */
	0x0000, "Access Associates",		/* Alegra */
	0x0000, "Anakin",			/* Easyl */
	0x0000, "ASDG",
	0x0000, "C.Ltd",
	0x0000, "Great Valley Products",
	0x0000, "MicroBotics",
	0x0000, "MicroForge",
	0x0000, "Micron Technologies",
	0x0000, "Phoenix",
	0x0000, "Progressive Peripherals",
	0x0000, "TecMar",
	0x0000, "Xebec",
	0xffff, "*unknown*"			/* must be last */
};

static char * ul_ON = "\033[4m";
static char * ul_OFF = "\033[0m";

static char * helptxt =
"%s displays all installed auto-configured expansion boards.  Known\n\
manufacturers and their codes are:\n\n\
  C O D E\n%s Dec%s %s Hex%s  %sMANUFACTURER%s\n";

static char * listtxt =
"\nZero codes are those presently unknown.  If you have updates or corrections,\
\nplease notify:\n\n\
UUCP:\tthad@cup.portal.com (OR) ...!sun!portal!cup.portal.com!thad\n\
UUCP:\tthad@well.UUCP\n\
BIX:\ttfloryan\n\
DELPHI:\tTHADF\n\
BBS:  *\tBBS-HT,   408/737-0900 (300,1200)      \"Thad Floryan\" or SYSOP\n\
BBS:\tBBS-JC,   415/961-7250 (300,1200,2400) \"Thad Floryan\" or SYSOP\n\
BBS:\tFAUG-BBS, 415/595-2479 (300,1200,2400) \"Thad Floryan\"\n\
BBS:\tAmiga OZ, 316/283-9210 (300,1200,2400) \"Thad Floryan\" or SYSOP\n\n\
      * = store BBS, California 6PM-8AM Mon-Fri, 5PM Sat-8AM Mon\n";

/***************************************************************************/

main(argc, argv)
    int argc;
    char *argv[];
{
    char *FindManufacturer();
    UWORD index = 0;
    struct ConfigDev *cd = 0L;

    if (argc > 1)
    {
	printf(helptxt, argv[0], ul_ON, ul_OFF, ul_ON, ul_OFF, ul_ON, ul_OFF);
	for (;;)
	{
	    printf("%4d %04x  %s\n",
			MfrTable[index].mc_Code,
			MfrTable[index].mc_Code,
			MfrTable[index].mc_Name);
	    if (MfrTable[++index].mc_Code == 0xffff) break;
	}
	printf(listtxt);
    }
    else
    {
	ExpansionBase = OpenLibrary(EXPANSIONNAME, 33L);
	if (ExpansionBase)
	{
	    while (cd = FindConfigDev(cd, -1L, -1L))
	    {
		printf("Board %d, Manufacturer %d [%04x] = %s\n",
			++index,
			cd->cd_Rom.er_Manufacturer,
			cd->cd_Rom.er_Manufacturer,
			FindManufacturer(cd->cd_Rom.er_Manufacturer));
		printf("\tMemory used   = %s\n",
			boardmemsize[(cd->cd_Rom.er_Type) & 7]);
		printf("\tProduct Code  = %d\n"   , cd->cd_Rom.er_Product);
		printf("\tSerial Number = %ld\n"  , cd->cd_Rom.er_SerialNumber);
		printf("\tBoard Address = 0x%lx\n", cd->cd_BoardAddr);
		printf("\tBoard Size    = %ld\n\n", cd->cd_BoardSize);
	    }
	    CloseLibrary(ExpansionBase);
	}
	else
	{
	    printf("?%s: Error, couldn't open library \"%s\"\n",
			argv[0],
			EXPANSIONNAME);
	}
    }
}

/*
 *	Returns the manufacturer's name given the CBM-assigned mfr code.
 *	This routine is presently a quick-n-dirty until a larger list of
 *	manufacturers and their codes is received.
 */

char *FindManufacturer(m_code)
    UWORD m_code;
{
    UWORD i, mfr;

    for (i = 0; ; i++)
    {
	mfr = MfrTable[i].mc_Code;
	if ( mfr == m_code || mfr == 0xffff )
	    return MfrTable[i].mc_Name;
    }
}
